""" Code for LAMOST data munging """

from __future__ import (absolute_import, division, print_function,)
import numpy as np
import scipy.optimize as opt
from scipy import interpolate 
import os
import sys
from cannon.helpers import Table
from cannon.dataset import Dataset
import matplotlib.pyplot as plt

# python 3 special
PY3 = sys.version_info[0] > 2
if not PY3:
    range = xrange

try:
    from astropy.io import fits as pyfits
except ImportError:
    import pyfits

def get_pixmask(file_in, wl, middle, flux, ivar):
    """ Return a mask array of bad pixels for one object's spectrum

    Bad pixels are defined as follows: fluxes or ivars are not finite, or 
    ivars are negative

    Major sky lines. 4046, 4358, 5460, 5577, 6300, 6363, 6863

    Where the red and blue wings join together: 5800-6000

    Read bad pix mask: file_in[0].data[4] is the ormask 

    Parameters
    ----------
    fluxes: ndarray
        flux array

    flux_errs: ndarray
        measurement uncertainties on fluxes

    Returns
    -------
    mask: ndarray, dtype=bool
        array giving bad pixels as True values
    """
    npix = len(wl)
    
    bad_flux = (~np.isfinite(flux)) # count: 0
    bad_err = (~np.isfinite(ivar)) | (ivar <= 0)
    # ivar == 0 for approximately 3-5% of pixels
    bad_pix_a = bad_err | bad_flux
    
    # LAMOST people: wings join together, 5800-6000 Angstroms
    wings = np.logical_and(wl > 5800, wl < 6000)
    # this is another 3-4% of the spectrum
    ormask = (file_in[0].data[4] >0)
    # ormask = (file_in[0].data[4] > 0)[middle]
    # ^ problematic...this is over a third of the spectrum!
    bad_pix_b = wings | ormask
    # bad_pix_b = wings

    spread = 3 # due to redshift
    skylines = np.array([4046, 4358, 5460, 5577, 6300, 6363, 6863])
    bad_pix_c = np.zeros(npix, dtype=bool)
    for skyline in skylines:
        badmin = skyline-spread
        badmax = skyline+spread
        bad_pix_temp = np.logical_and(wl > badmin, wl < badmax)
        bad_pix_c[bad_pix_temp] = True
    # 34 pixels

    bad_pix_ab = bad_pix_a | bad_pix_b
    bad_pix = bad_pix_ab | bad_pix_c

    return bad_pix


def load_spectra(data_dir, filenames):
    """
    Extracts spectra (wavelengths, fluxes, fluxerrs) from apogee fits files

    Returns
    -------
    IDs: list of length nstars
        stellar IDs
    
    wl: numpy ndarray of length npixels
        rest-frame wavelength vector

    fluxes: numpy ndarray of shape (nstars, npixels)
        training set or test set pixel intensities

    ivars: numpy ndarray of shape (nstars, npixels)
        inverse variances, parallel to fluxes
        
    SNRs: numpy ndarray of length nstars
    """
    print("Loading spectra from directory %s" %data_dir)
    files = list(sorted(filenames))
    files = np.array(files)
    nstars = len(files)

    npix = np.zeros(nstars)
    for jj, fits_file in enumerate(files):
        file_in = pyfits.open("%s/%s" %(data_dir, fits_file))
        grid_all = np.array(file_in[0].data[2])
        if jj == 0:
            # all stars do NOT start out on the same wavelength grid
            # grid_all = np.array(file_in[0].data[2])
            # some spectra will end up with different pixels
            # because of the wavelength correction. so do this to ensure
            # that the interpolation never extrapolates...
            # values determined by experimentation, may change later
            # middle = np.logical_and(grid_all > 3705, grid_all < 9091)
            # only lost 10 pixels here
            # now, add on 200 pixels at the beginning to save the Ca H&K
            middle = np.logical_and(grid_all > 3905, grid_all < 9000)
            grid = grid_all[middle]
            npixels = len(grid) 
            SNRs = np.zeros(nstars, dtype=float)   
            fluxes = np.zeros((nstars, npixels), dtype=float)
            ivars = np.zeros(fluxes.shape, dtype=float)
            #badpixs = np.zeros((nstars, len(grid_all)), dtype=bool)
        flux = np.array(file_in[0].data[0])
        npix[jj] = len(flux)
        ivar = np.array((file_in[0].data[1]))
        # identify bad pixels PRIOR to shifting, so that the sky lines
        # don't move around
        badpix = get_pixmask(file_in, grid_all, middle, flux, ivar)
        flux = np.ma.array(flux, mask=badpix)
        ivar = np.ma.array(ivar, mask=badpix)
        SNRs[jj] = np.ma.median(flux*ivar**0.5)
        ivar = np.ma.filled(ivar, fill_value=0.)
        # correct for radial velocity of star
        redshift = file_in[0].header['Z']
        wlshift = redshift*grid_all
        wl = grid_all - wlshift
        # resample onto a common grid
        flux_rs = (interpolate.interp1d(wl, flux))(grid)
        ivar_rs = (interpolate.interp1d(wl, ivar))(grid)
        fluxes[jj,:] = flux_rs
        ivars[jj,:] = ivar_rs

    print("Spectra loaded")
    return grid, fluxes, ivars


def load_labels(label_file):
    """ Extracts training labels from file.

    Assumes that first row is # then label names, first col is # then 
    filenames, remaining values are floats and user wants all the labels.
    """
    print("Loading reference labels from file %s" %label_file)
    data = Table(label_file)
    data.sort('id')
    label_names = data.keys()[1:] # ignore id
    nlabels = len(label_names)
    print('%s labels:' %nlabels)
    print(label_names)
    labels = np.array([data[k] for k in label_names], dtype=float).T
    return data['id'], labels 



""" Make a single model class to rule them all """
from .dataset import Dataset
from .train_model import train_model as _train_model
from .infer_labels import infer_labels
from .spectral_model import diagnostics as _diagnostics
from .helpers.triangle import corner
import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy

LARGE = 200.
SMALL = 1. / LARGE

class CannonModel(object):
    def __init__(self, dataset, order):
        if not isinstance(dataset, Dataset):
            txt = 'Expecting a Dataset instance, got {0}'
            raise TypeError(txt.format(type(dataset)))
        self.dataset = dataset
        self._model = None
        self.order = order

    @property
    def model(self):
        """ return the model definition or raise an error if not trained """
        if self._model is None:
            raise RuntimeError('Model not trained')
        else:
            return self._model

    def train(self, *args, **kwargs):
        """ Train the model """
        self._model = _train_model(self.dataset)

    def diagnostics(self):
        """Run a set of diagnostics on the model.

        Plot the 0th order coefficients as the baseline spectrum.
        Overplot the continuum pixels.

        Plot each label's leading coefficient as a function of wavelength.
        Color-code by label.

        Histogram of the chi squareds of the fits.
        Dotted line corresponding to DOF = npixels - nlabels

        Parameters
        ----------
        contpix: str
            continuum pixel definition file
        """
        _model_diagnostics(self.dataset, self.model)

    def infer_labels(self, dataset):
        """
        Uses the model to solve for labels of the test set.

        Parameters
        ----------
        test_set: Dataset
            dataset that needs label inference

        Returns
        -------
        test_set: Dataset
            same dataset as the input value with updated label_vals attribute

        covs_all:
            covariance matrix of the fit
        """
        return infer_labels(self.model, dataset)

    def draw_spectra(self, dataset):
        """
        Parameters
        ----------
        test_set: Dataset
            dataset that needs label inference

        Returns
        -------
        cannon_set: Dataset
            same dataset as the input value with updated fluxes and variances

        """
        coeffs_all, covs, scatters, red_chisqs, pivots, label_vector = self.model
        nstars = len(dataset.test_SNR)
        cannon_fluxes = np.zeros(dataset.test_flux.shape)
        cannon_ivars = np.zeros(dataset.test_ivar.shape)
        for i in range(nstars):
            x = label_vector[:,i,:]
            spec_fit = np.einsum('ij, ij->i', x, coeffs_all)
            cannon_fluxes[i,:] = spec_fit
            bad = dataset.test_ivar[i,:] == SMALL
            cannon_ivars[i,:][~bad] = 1. / scatters[~bad] ** 2
        cannon_set = deepcopy(dataset)
        cannon_set.test_flux = cannon_fluxes
        cannon_set.test_ivar = cannon_ivars

        return cannon_set

    def split_array(self, array, num):
        avg = len(array) / float(num)
        out = []
        last = 0.0
        while last < len(array):
            out.append(array[int(last):int(last+avg)])
            last += avg
        return out

    def diagnostics(
            self, baseline_spec_plot_name = "baseline_spec_with_cont_pix",
            leading_coeffs_plot_name = "leading_coeffs.png",
            chisq_dist_plot_name = "modelfit_chisqs.png"):
        """Run a set of diagnostics on the model.

        Plot the 0th order coefficients as the baseline spectrum.
        Overplot the continuum pixels.

        Plot each label's leading coefficient as a function of wavelength.
        Color-code by label.

        Histogram of the chi squareds of the fits.
        Dotted line corresponding to DOF = npixels - nlabels

        Parameters
        ----------
        """
        dataset = self.dataset
        model = self.model

        contmask = dataset.contmask
        lams = dataset.wl
        label_names = dataset.get_plotting_labels()
        coeffs_all, covs, scatters, chisqs, pivots, label_vector = model
        npixels = len(lams)
        nlabels = len(pivots)

        if contmask != None:
            # Baseline spectrum with continuum
            baseline_spec = coeffs_all[:,0]
            bad = np.round(baseline_spec,5) == 0
            baseline_spec = np.ma.array(baseline_spec, mask=bad)
            lams = np.ma.array(lams, mask=bad)

            # Continuum pixels
            contpix_lambda = lams[contmask]
            y = baseline_spec[contmask]

            # Split into ten segments
            nseg = 10
            lams_seg = self.split_array(lams.compressed(), nseg)
            xmins = [] 
            xmaxs = [] 
            for seg in lams_seg:
                xmins.append(seg[0])
                xmaxs.append(seg[-1])

            for i in range(nseg):
                fig, axarr = plt.subplots(2, sharex=True)
                plt.xlabel(r"Wavelength $\lambda (\AA)$")
                plt.xlim(xmins[i], xmaxs[i])
                ax = axarr[0]
                ax.step(lams, baseline_spec, where='mid', c='k', linewidth=0.3,
                        label=r'$\theta_0$' + "= the leading fit coefficient")
                ax.scatter(contpix_lambda, y, s=1, color='r',label="continuum pixels")
                ax.legend(loc='lower right', prop={'family':'serif', 'size':'small'})
                ax.set_title("Baseline Spectrum with Continuum Pixels")
                ax.set_ylabel(r'$\theta_0$')
                ax = axarr[1]
                ax.step(lams, baseline_spec, where='mid', c='k', linewidth=0.3,
                     label=r'$\theta_0$' + "= the leading fit coefficient")
                ax.scatter(contpix_lambda, y, s=1, color='r',label="continuum pixels")
                ax.set_title("Baseline Spectrum with Continuum Pixels, Zoomed")
                ax.legend(loc='upper right', prop={'family':'serif', 'size':'small'})
                ax.set_ylabel(r'$\theta_0$')
                ax.set_ylim(0.95, 1.05)
                print("Diagnostic plot: fitted 0th order spec, cont pix overlaid.")
                print("Saved as %s_%s.png" % (baseline_spec_plot_name, i))
                plt.savefig(baseline_spec_plot_name + "_%s" %i)
                plt.close()

        # Leading coefficients for each label & scatter
        # Scale coefficients so that they can be overlaid on the same plot
        c = ['r', 'g', 'b', 'k']
        stds = np.array([np.std(coeffs_all[:, i + 1]) for i in range(nlabels)])
        pivot_std = max(stds)
        ratios = np.round(pivot_std / stds, -1)  # round to the nearest 10
        ratios[ratios == 0] = 1
        fig, axarr = plt.subplots(2, sharex=True)
        plt.xlabel(r"Wavelength $\lambda (\AA)$")
        plt.xlim(np.ma.min(lams), np.ma.max(lams))
        ax = axarr[0]
        ax.set_ylabel("Leading coefficient " + r"$\theta_i$")
        ax.set_title("First-Order Fit Coefficients for Labels")

        first_order = np.zeros((npixels, nlabels))
        lbl = r'$\theta_{0:d}$=coeff for ${1:s}$ * ${2:d}$'
        for i in range(nlabels):
            coeffs = coeffs_all[:,i+1] * ratios[i]
            first_order[:,i] = coeffs
            ax.step(lams, coeffs, where='mid', linewidth=0.5, 
                    label=lbl.format(i+1, label_names[i], int(ratios[i])), c=c[i])
        box = ax.get_position()
        ax.set_position(
                [box.x0, box.y0 + box.height*0.1, box.width, box.height*0.9])
        ax.legend(
                bbox_to_anchor=(0., -.2, 1., .102), loc=3, ncol=3, 
                mode="expand", prop={'family':'serif', 'size':'small'})
        ax = axarr[1]
        ax.set_ylabel("scatter")
        ax.set_title("Scatter of Fit")
        ax.step(lams, scatters, where='mid', c='k', linewidth=0.7)
        fig.tight_layout(pad=2.0, h_pad=4.0)
        print("Diagnostic plot: leading coeffs and scatters across wavelength.")
        print("Saved as %s" %leading_coeffs_plot_name)
        fig.savefig(leading_coeffs_plot_name)
        plt.close(fig)

        # triangle plot of the higher-order coefficients
        labels = [r"$%s$" % l for l in label_names]
        fig = corner(first_order, labels=labels, show_titles=True,
                     title_args = {"fontsize":12})
        filename = "leading_coeffs_triangle.png"
        fig.savefig(filename)
        plt.close(fig)

        # Histogram of the chi squareds of ind. stars
        plt.hist(np.sum(chisqs, axis=0), color='lightblue', alpha=0.7)
        dof = len(lams) - coeffs_all.shape[1]   # for one star
        plt.axvline(x=dof, c='k', linewidth=2, label="DOF")
        plt.legend()
        plt.title("Distribution of " + r"$\chi^2$" + " of the Model Fit")
        plt.ylabel("Count")
        plt.xlabel(r"$\chi^2$" + " of Individual Star")
        print("Diagnostic plot: histogram of the red chi squareds of the fit")
        print("Saved as %s" %chisq_dist_plot_name)
        plt.savefig(chisq_dist_plot_name)
        plt.close()

    # convenient namings to match existing packages
    predict = infer_labels
    fit = train



"""This runs Step 1 of The Cannon:
    uses the training set to solve for the best-fit model."""

from __future__ import (absolute_import, division, print_function, unicode_literals)
import numpy as np
import matplotlib.pyplot as plt
from .helpers.compatibility import range, map
from .helpers.triangle import corner

def do_one_regression_at_fixed_scatter(lams, fluxes, ivars, lvec, scatter):
    """
    Parameters
    ----------
    lams: numpy ndarray, shape npixels
        the common wavelength array

    fluxes: numpy ndarray, shape (nstars, npixels)
        flux values for all pixels of all stars
        
    ivars: numpy ndarray, shape (nstars, npixels)
        inverse variance values for all pixels of all stars

    lvec: numpy ndarray
        the label vector

    scatter: numpy ndarray, shape npixels
        fitted scatter values

    Returns
    ------
    coeff: ndarray
        coefficients of the fit

    lTCinvl: ndarray
        inverse covariance matrix for fit coefficients
    
    chi: float
        chi-squared at best fit
    
    logdet_Cinv: float
        inverse of the log determinant of the cov matrix
    """
    #sig2 = 100**2*np.ones(len(ivars))
    #mask = ivars != 0
    #sig2[mask] = 1. / ivars[mask]
    sig2 = 1. / ivars
    Cinv = 1. / (sig2 + scatter**2)
    lTCinvl = np.dot(lvec.T, Cinv[:, None] * lvec)
    lTCinvf = np.dot(lvec.T, Cinv * fluxes)
    try:
        coeff = np.linalg.solve(lTCinvl, lTCinvf)
    except np.linalg.linalg.LinAlgError:
        print("np.linalg.linalg.LinAlgError, do_one_regression_at_fixed_scatter")
        print(lTCinvl, lTCinvf, lams, fluxes)
    if not np.all(np.isfinite(coeff)):
        raise RuntimeError('something is wrong with the coefficients')
    chi = np.sqrt(Cinv) * (fluxes - np.dot(lvec, coeff))
    logdet_Cinv = np.sum(np.log(Cinv))
    return (coeff, lTCinvl, chi, logdet_Cinv)


def do_one_regression(lams, fluxes, ivars, lvec):
    """
    Optimizes to find the scatter associated with the best-fit model.

    This scatter is the deviation between the observed spectrum and the model.
    It is wavelength-independent, so we perform this at a single wavelength.

    Input
    -----
    lams: numpy ndarray, shape (npixels)
        the common wavelength array

    fluxes: numpy ndarray, shape (nstars)

    ivars: numpy ndarray, shape (nstars)

    lvec = numpy ndarray 
        the label vector

    Output
    -----
    output of do_one_regression_at_fixed_scatter
    """
    ln_scatter_vals = np.arange(np.log(0.0001), 0., 0.5)
    # minimize over the range of scatter possibilities
    chis_eval = np.zeros_like(ln_scatter_vals)
    for jj, ln_scatter_val in enumerate(ln_scatter_vals):
        coeff, lTCinvl, chi, logdet_Cinv = \
            do_one_regression_at_fixed_scatter(lams, fluxes, ivars, lvec,
                                               scatter=np.exp(ln_scatter_val))
        chis_eval[jj] = np.sum(chi*chi) - logdet_Cinv
    if np.any(np.isnan(chis_eval)):
        best_scatter = np.exp(ln_scatter_vals[-1])
        _r = do_one_regression_at_fixed_scatter(lams, fluxes, ivars, lvec,
                                                scatter=best_scatter)
        return _r + (best_scatter, )
    lowest = np.argmin(chis_eval)
    if (lowest == 0) or (lowest == len(ln_scatter_vals) - 1):
        best_scatter = np.exp(ln_scatter_vals[lowest])
        _r = do_one_regression_at_fixed_scatter(lams, fluxes, ivars, lvec,
                                                scatter=best_scatter)
        return _r + (best_scatter, )
    ln_scatter_vals_short = ln_scatter_vals[np.array(
        [lowest-1, lowest, lowest+1])]
    chis_eval_short = chis_eval[np.array([lowest-1, lowest, lowest+1])]
    z = np.polyfit(ln_scatter_vals_short, chis_eval_short, 2)
    fit_pder = np.polyder(z)
    best_scatter = np.exp(np.roots(fit_pder)[0])
    _r = do_one_regression_at_fixed_scatter(lams, fluxes, ivars, lvec,
                                            scatter=best_scatter)
    return _r + (best_scatter, )


def get_lvec(label_vals, pivots):
    """
    Constructs a label vector for an arbitrary number of labels
    Assumes that our model is quadratic in the labels

    Parameters
    ----------
    label_vals: numpy ndarray, shape (nstars, nlabels)
    pivots: array corresponding to the mean of the label_vals

    Returns
    -------
    lvec_full: numpy ndarray
        label vector
    """
    nlabels = label_vals.shape[1]
    nstars = label_vals.shape[0]
    # specialized to second-order model
    linear_offsets = label_vals - pivots
    quadratic_offsets = np.array([np.outer(m, m)[np.triu_indices(nlabels)]
                                  for m in (linear_offsets)])
    ones = np.ones((nstars, 1))
    lvec = np.hstack((ones, linear_offsets, quadratic_offsets))
    return lvec


def train_model(dataset):
    """
    This determines the coefficients of the model using the training data

    Parameters
    ----------
    reference_set: Dataset

    Returns
    -------
    model, which consists of:
   
    coeffs: ndarray, (npixels, nstars, 15)
        the model coefficients

    covs: ndarray
        the covariance matrix

    scatter:
        scatter values

    all_chisqs:

    pivots:
        the pivot values

    lvec_full:
        the label vector
    """
    print("Training model...")
    label_names = dataset.get_plotting_labels()
    label_vals = dataset.tr_label
    lams = dataset.wl
    npixels = len(lams)
    fluxes = dataset.tr_flux
    ivars = dataset.tr_ivar
    pivots = np.mean(label_vals, axis=0)
    lvec = get_lvec(label_vals, pivots)
    lvec_full = np.array([lvec,] * npixels)

    # Perform REGRESSIONS
    fluxes = fluxes.swapaxes(0,1)  # for consistency with lvec_full
    ivars = ivars.swapaxes(0,1)
    # one per pix
    blob = list(map(do_one_regression, lams, fluxes, ivars, lvec_full))
    coeffs = np.array([b[0] for b in blob])
    covs = np.array([np.linalg.inv(b[1]) for b in blob])
    chis = np.array([b[2] for b in blob])
    scatters = np.array([b[4] for b in blob])

    # Calc chi sq
    all_chisqs = chis*chis
    model = coeffs, covs, scatters, all_chisqs, pivots, lvec_full
    print("Done training model")
    return model


def split_array(array, num):
    avg = len(array) / float(num)
    out = []
    last = 0.0
    while last < len(array):
        out.append(array[int(last):int(last+avg)])
        last += avg
    return out



import numpy as np
from functools import partial
import matplotlib.pyplot as plt
import scipy.optimize as opt

""" Performs continuum normalization on Cannon input spectra. """

LARGE = 200.
SMALL = 1. / LARGE

def partial_func(func, *args, **kwargs):
    def wrap(x, *p):
        return func(x, p, **kwargs)
    return wrap


def cont_func(x, p, L, y):
    """ Return the fitting function for the continuum.

    Parameters
    ----------
    x: float or np.array
    p: ndarray
        function coefficients. first element L is not fitted for.

    Returns
    -------
    func: float
        function evaluated for the input x
    """
    N = int(len(p)/2)
    n = np.linspace(0, N, N+1, dtype=int)
    k = n*np.pi/L
    baseline = 0 #if you were fitting a flat spectrum...
    #baseline = y[x]
    func = 0
    for n in range(0, N):
        func += p[2*n]*np.sin(k[n]*x)+p[2*n+1]*np.cos(k[n]*x)
    return func
    #return baseline+func


def fit_cont(fluxes, ivars, contmask, deg, ffunc):
    """ Fit a continuum to a continuous segment of spectra.

    Fit a function of sines and cosines with specified degree.
    """
    print("order: %s" %deg)
    nstars = fluxes.shape[0]
    npixels = fluxes.shape[1]
    cont = np.zeros(fluxes.shape)
    
    for jj in range(nstars):
        # Fit continuum to cont pixels
        flux = fluxes[jj,:]
        ivar = ivars[jj,:]
        pix = np.arange(0, npixels)
        y = flux[contmask]
        x = pix[contmask]
        yivar = ivar[contmask]
        yivar[yivar == 0] = SMALL**2   
        p0 = np.ones(deg*2) # one for cos, one for sin
        L = max(x)-min(x)
        pcont_func = partial_func(cont_func, L=L, y=flux)
        # It should not be possible to have a cont pix that's also
        # a bad pixel, because bad means flux_err==0, and flux_err==0
        # should also correspond to flux==0. If flux==0 at this pixel,
        # it should throw off the median(flux) and ivar(flux) cuts.
        # bad = yivar == 0 #| yivar == SMALL 
        # yivar = np.ma.array(yivar, mask=bad)
        
        if ffunc=="sinusoid": 
            popt, pcov = opt.curve_fit(pcont_func, x, y, p0=p0, 
                                       sigma=1./np.sqrt(yivar))
        elif ffunc=="chebyshev":
            fit = np.polynomial.chebyshev.Chebyshev.fit(x=x,y=y,w=yivar,deg=deg)

        for element in pix:
            if ffunc=="sinusoid":
                cont[jj,element] = cont_func(element, popt, L=L, y=flux)
            elif ffunc=="chebyshev":
                cont[element] = fit(element)

    return cont


def fit_cont_regions(fluxes, ivars, contmask, deg, ranges, ffunc):
    print("taking spectra in %s regions" %len(ranges))
    nstars = fluxes.shape[0]
    npixels = fluxes.shape[1]
    cont = np.zeros(fluxes.shape)
    for chunk in ranges:
        start = chunk[0]
        stop = chunk[1]
        if ffunc=="chebyshev":
            output = fit_cont(fluxes[:,start:stop],
                              ivars[:,start:stop],
                              contmask[start:stop], deg=deg, ffunc="chebyshev")
        elif ffunc=="sinusoid":
            output = fit_cont(fluxes[:,start:stop],
                              ivars[:,start:stop],
                              contmask[start:stop], deg=deg, ffunc="sinusoid")
        cont[:,start:stop] = output
    return cont


def cont_norm(fluxes, ivars, cont):
    """ Continuum-normalize a continuous segment of spectra.

    Fit has already been performed.

    Parameters
    ----------
    fluxes: numpy ndarray 
        pixel intensities
    ivars: numpy ndarray 
        inverse variances, parallel to fluxes
    contmask: boolean mask
        True indicates that pixel is continuum
    deg: (optional) int
        degree of fit, corresponds to # of sines or # of cosines

    Returns
    -------
    norm_fluxes: numpy ndarray
        normalized pixel intensities
    norm_ivars: numpy ndarray
        rescaled inverse variances
    """
    nstars = fluxes.shape[0]
    norm_fluxes = np.zeros(fluxes.shape)
    norm_ivars = np.zeros(ivars.shape)
    for jj in range(nstars):
        bad = (ivars[jj,:] == 0)
        norm_fluxes[jj,:] = fluxes[jj,:]/cont[jj,:]
        norm_ivars[jj,:] = cont[jj,:]**2 * ivars[jj,:]
        norm_fluxes[jj,:][bad] = 1.
        norm_ivars[jj,:][bad] = SMALL**2
    return norm_fluxes, norm_ivars 


def weighted_median(values, weights, quantile):
    sindx = np.argsort(values)
    cvalues = 1. * np.cumsum(weights[sindx])
    cvalues = cvalues / cvalues[-1]
    foo = sindx[cvalues > quantile]
    if len(foo) == 0:
        return values[0]
    indx = foo[0]
    return values[indx]


def cont_norm_q(wl, fluxes, ivars, q, delta_lambda):
    print("contnorm.py: continuum norm using running quantile")
    norm_fluxes = np.zeros(fluxes.shape)
    norm_ivars = np.zeros(ivars.shape)
    cont = np.zeros(fluxes.shape)
    nstars = fluxes.shape[0]
    for jj in range(nstars):
        print("cont_norm_q(): working on star %s" %jj) 
        flux = fluxes[jj,:]
        ivar = ivars[jj,:]
        for ll, lam in enumerate(wl):
            indx = (np.where(abs(wl-lam) < delta_lambda))[0]
            flux_cut = flux[indx]
            ivar_cut = ivar[indx]
            cont[jj,ll] = weighted_median(flux_cut, ivar_cut, q)
    for jj in range(nstars):
        norm_fluxes[jj,:] = fluxes[jj,:]/cont[jj,:]
        norm_ivars[jj,:] = cont[jj,:]**2 * ivars[jj,:]
    return norm_fluxes, norm_ivars


def cont_norm_regions(fluxes, ivars, cont, ranges):
    print("taking spectra in %s regions" %len(ranges))
    nstars = fluxes.shape[0]
    norm_fluxes = np.zeros(fluxes.shape)
    norm_ivars = np.zeros(ivars.shape)
    for chunk in ranges:
        start = chunk[0]
        stop = chunk[1]
        output = cont_norm(fluxes[:,start:stop],
                           ivars[:,start:stop],
                           cont[:,start:stop])
        norm_fluxes[:,start:stop] = output[0]
        norm_ivars[:,start:stop] = output[1]
    for jj in range(nstars):
        bad = (norm_ivars[jj,:] == 0.)
        #bad = (norm_ivars[jj,:] == SMALL**2)
        norm_fluxes[jj,:][bad] = 0.
        norm_ivars[jj,:][bad] = SMALL**2
    return norm_fluxes, norm_ivars




from __future__ import (absolute_import, division, print_function, unicode_literals)
import numpy as np
import os
import random
from .helpers.triangle import corner
import matplotlib.pyplot as plt
from matplotlib import colorbar
from copy import deepcopy

LARGE = 200.
SMALL = 1. / LARGE

def draw_spectra(model, dataset):
    """ Generate a set of spectra given the model and test set labels 

    Parameters
    ----------
    model:

    test_set:

    Returns
    -------
    cannon_set
    """
    
    coeffs_all, covs, scatters, red_chisqs, pivots, label_vector = model
    nstars = len(dataset.test_SNRs)
    cannon_fluxes = np.zeros(dataset.test_fluxes.shape)
    cannon_ivars = np.zeros(dataset.test_ivars.shape)
    for i in range(nstars):
        x = label_vector[:,i,:]
        spec_fit = np.einsum('ij, ij->i', x, coeffs_all)
        cannon_fluxes[i,:] = spec_fit
        bad = dataset.test_ivars[i,:] == SMALL
        cannon_ivars[i,:][~bad] = 1. / scatters[~bad] ** 2
    cannon_set = deepcopy(dataset)
    cannon_set.test_fluxes = cannon_fluxes
    cannon_set.test_ivars = cannon_ivars

    return cannon_set


def diagnostics(cannon_set, dataset, model):
    """ Run diagnostics on the fitted spectra

    Parameters
    ----------
    cannon_set:

    test_set:

    model:
    """
    #overlay_spectra(cannon_set, dataset, model)
    residuals(cannon_set, dataset)


def triangle_pixels(cannon_set):
    """ Make a triangle plot for the pixel fluxes, to see correlation
    
    Parameters
    ----------
    cannon_set
    """
    
    fluxes = cannon_set.test_fluxes
    corner(fluxes)


def overlay_spectra(cannon_set, dataset, model):
    """ Run a series of diagnostics on the fitted spectra 

    Parameters
    ----------

    cannon_set:

    test_set:

    model:
    """
    
    coeffs_all, covs, scatters, chisqs, pivots, label_vector = model
    # Overplot original spectra with best-fit spectra
    res = dataset.test_fluxes-cannon_set.test_fluxes
    bad_ivar = np.std(res, axis=0) <= 1e-5
    os.system("mkdir SpectrumFits")
    print("Overplotting spectra for ten random stars")
    lambdas = dataset.wl
    npix = len(lambdas)
    nstars = cannon_set.test_fluxes.shape[0]
    pickstars = []
    for i in range(10):
        pickstars.append(random.randrange(0, nstars-1))
    for i in pickstars:
        print("Star %s" % i)
        ID = cannon_set.test_IDs[i]
        spec_orig = dataset.test_fluxes[i,:]
        bad_flux = (spec_orig == 0) | (spec_orig == 1)  # unique to star
        bad = (bad_ivar | bad_flux)
        lambdas = np.ma.array(lambdas, mask=bad, dtype=float)
        npix = len(lambdas.compressed())
        spec_orig = np.ma.array(dataset.test_fluxes[i,:], mask=bad)
        spec_fit = np.ma.array(cannon_set.test_fluxes[i,:], mask=bad)
        ivars_orig = np.ma.array(dataset.test_ivars[i,:], mask=bad)
        ivars_fit = np.ma.array(cannon_set.test_ivars[i,:], mask=bad)
        #fig, axarr = plt.subplots(2)
        #ax1 = axarr[0]
        #ax1.scatter(lambdas, spec_orig, label="Orig Spec",
        #            c=1. / np.sqrt(ivars_orig))
        red_chisq = np.sum(chisqs[:,i], axis=0) / (npix - coeffs_all.shape[1])
        red_chisq = np.round(red_chisq, 2)
        fig,axarr = plt.subplots(2)
        ax1 = axarr[0]
        im = ax1.scatter(lambdas, spec_orig, label="Orig Spec",
                         c=1 / np.sqrt(ivars_orig))
        ax1.scatter(lambdas, spec_fit, label="Cannon Spec", c='r')
        ax1.errorbar(lambdas, spec_fit, yerr=1/np.sqrt(ivars_fit), fmt='ro')
        ax1.set_xlabel(r"Wavelength $\lambda (\AA)$")
        ax1.set_ylabel("Normalized flux")
        ax1.set_title("Spectrum Fit: %s" % ID)
        ax1.set_title("Spectrum Fit")
        ax1.legend(loc='lower center', fancybox=True, shadow=True)
        ax2 = axarr[1]
        ax2.scatter(spec_orig, spec_fit, c=1/np.sqrt(ivars_orig))
        ax2.errorbar(spec_orig, spec_fit, yerr=1 / np.sqrt(ivars_fit),
                     ecolor='k', fmt="none")
        fig.subplots_adjust(right=0.8)
        cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
        fig.colorbar(
                im, cax=cbar_ax,
                label="Uncertainties on the Fluxes from the Original Spectrum")
        xlims = ax2.get_xlim()
        ylims = ax2.get_ylim()
        lims = [np.min([xlims, ylims]), np.max([xlims, ylims])]
        ax2.plot(lims, lims, 'k-', alpha=0.75)
        textstr = "Red Chi Sq: %s" % red_chisq
        props = dict(boxstyle='round', facecolor='palevioletred', alpha=0.5)
        ax2.text(0.05, 0.95, textstr, transform=ax2.transAxes, fontsize=14,
                 verticalalignment='top', bbox=props)
        ax2.set_xlim(xlims)
        ax2.set_ylim(ylims)
        ax2.set_xlabel("Orig Fluxes")
        ax2.set_ylabel("Fitted Fluxes")
        filename = "Star%s.png" % i
        print("Saved as %s" % filename)
        fig.savefig("SpectrumFits/" + filename)
        plt.close(fig)


def residuals(cannon_set, dataset):
    """ Stack spectrum fit residuals, sort by each label. Include histogram of
    the RMS at each pixel.

    Parameters
    ----------
    cannon_set:

    test_set:
    """
    print("Stacking spectrum fit residuals")
    res = dataset.test_fluxes - cannon_set.test_fluxes
    bad = dataset.test_ivars == SMALL**2
    err = np.zeros(len(dataset.test_ivars))
    err = np.sqrt(1. / dataset.test_ivars + 1. / cannon_set.test_ivars)
    res_norm = res / err
    res_norm = np.ma.array(res_norm,
                           mask=(np.ones_like(res_norm) *
                                 (np.std(res_norm,axis=0) == 0)))
    res_norm = np.ma.compress_cols(res_norm)

    for i in range(len(cannon_set.get_plotting_labels())):
        label_name = cannon_set.get_plotting_labels()[i]
        print("Plotting residuals sorted by %s" % label_name)
        label_vals = cannon_set.tr_label_vals[:,i]
        sorted_res = res_norm[np.argsort(label_vals)]
        mu = np.mean(sorted_res.flatten())
        sigma = np.std(sorted_res.flatten())
        left, width = 0.1, 0.65
        bottom, height = 0.1, 0.65
        bottom_h = left_h = left+width+0.1
        rect_scatter = [left, bottom, width, height]
        rect_histx = [left, bottom_h, width, 0.1]
        rect_histy = [left_h, bottom, 0.1, height]
        plt.figure()
        axScatter = plt.axes(rect_scatter)
        axHistx = plt.axes(rect_histx)
        axHisty = plt.axes(rect_histy)
        im = axScatter.imshow(sorted_res, cmap=plt.cm.bwr_r,
                              interpolation="nearest", vmin=mu - 3. * sigma,
                              vmax=mu + 3. * sigma, aspect='auto',
                              origin='lower', extent=[0, len(dataset.wl),
                                                      min(label_vals),
                                                      max(label_vals)])
        cax, kw = colorbar.make_axes(axScatter.axes, location='bottom')
        plt.colorbar(im, cax=cax, orientation='horizontal')
        axScatter.set_title(
                r"Spectral Residuals Sorted by ${0:s}$".format(label_name))
        axScatter.set_xlabel("Pixels")
        axScatter.set_ylabel(r"$%s$" % label_name)
        axHisty.hist(np.std(res_norm,axis=1)[~np.isnan(np.std(res_norm, axis=1))], orientation='horizontal', range=[0,2])
        axHisty.axhline(y=1, c='k', linewidth=3, label="y=1")
        axHisty.legend(bbox_to_anchor=(0., 0.8, 1., .102),
                       prop={'family':'serif', 'size':'small'})
        axHisty.text(1.0, 0.5, "Distribution of Stdev of Star Residuals",
                     verticalalignment='center', transform=axHisty.transAxes,
                     rotation=270)
        axHisty.set_ylabel("Standard Deviation")
        start, end = axHisty.get_xlim()
        axHisty.xaxis.set_ticks(np.linspace(start, end, 3))
        axHisty.set_xlabel("Number of Stars")
        axHisty.xaxis.set_label_position("top")
        axHistx.hist(np.std(res_norm, axis=0)[~np.isnan(np.std(res_norm, axis=0))], range=[0.8,1.1])
        axHistx.axvline(x=1, c='k', linewidth=3, label="x=1")
        axHistx.set_title("Distribution of Stdev of Pixel Residuals")
        axHistx.set_xlabel("Standard Deviation")
        axHistx.set_ylabel("Number of Pixels")
        start, end = axHistx.get_ylim()
        axHistx.yaxis.set_ticks(np.linspace(start, end, 3))
        axHistx.legend()
        filename = "residuals_sorted_by_label_%s.png" % i
        plt.savefig(filename)
        print("File saved as %s" % filename)
        plt.close()

    # Auto-correlation of mean residuals
    print("Plotting Auto-Correlation of Mean Residuals")
    mean_res = res_norm.mean(axis=0)
    autocorr = np.correlate(mean_res, mean_res, mode="full")
    pkwidth = int(len(autocorr)/2-np.argmin(autocorr))
    xmin = int(len(autocorr)/2)-pkwidth
    xmax = int(len(autocorr)/2)+pkwidth
    zoom_x = np.linspace(xmin, xmax, len(autocorr[xmin:xmax]))
    fig, axarr = plt.subplots(2)
    axarr[0].plot(autocorr)
    axarr[0].set_title("Autocorrelation of Mean Spectral Residual")
    axarr[0].set_xlabel("Lag (# Pixels)")
    axarr[0].set_ylabel("Autocorrelation")
    axarr[1].plot(zoom_x, autocorr[xmin:xmax])
    axarr[1].set_title("Central Peak, Zoomed")
    axarr[1].set_xlabel("Lag (# Pixels)")
    axarr[1].set_ylabel("Autocorrelation")
    filename = "residuals_autocorr.png"
    plt.savefig(filename)
    print("saved %s" % filename)
    plt.close()



from __future__ import (absolute_import, division, print_function)
import numpy as np
import matplotlib.pyplot as plt
from .helpers.triangle import corner
from helpers import Table
import sys
from .find_continuum_pixels import find_contpix, find_contpix_regions
from .continuum_normalization import fit_cont, fit_cont_regions, cont_norm, cont_norm_regions, cont_norm_q

PY3 = sys.version_info[0] > 2

if PY3:
    basestring = (str, bytes)
else:
    basestring = (str, unicode)


class Dataset(object):
    """A class to represent a Dataset of stellar spectra and labels.

    Initialize this object after performing the munging necessary 
    for making data "Cannonizable."
    """

    def __init__(self, wl, tr_flux, tr_ivar, tr_label, test_flux, test_ivar):
        print("Loading dataset")
        print("This may take a while...")
        self.wl = wl
        self.tr_flux = tr_flux
        self.tr_ivar = tr_ivar
        self.tr_label = tr_label
        self.test_flux = test_flux
        self.test_ivar = test_ivar
        self.ranges = None
        
        # calculate SNR
        self.tr_SNR = np.array(
                [self._SNR(*s) for s in zip(tr_flux, tr_ivar)])
        self.test_SNR = np.array(
                [self._SNR(*s) for s in zip(test_flux, test_ivar)])


    def _SNR(self, flux, ivar):
        """ Calculate the SNR of a spectrum, ignoring bad pixels

        Parameters
        ----------
        flux: 1D array of flux values
        ivar: 1D array of ivar values

        Returns
        -------
        SNR: float
        """
        bad = ivar == 0
        flux = np.ma.array(flux, mask=bad)
        ivar = np.ma.array(ivar, mask=bad)
        SNR = float(np.ma.median(flux*ivar**0.5).data)
        return SNR  


    def set_label_names(self, names):
        self._label_names = names


    def get_plotting_labels(self):
        if self._label_names is None:
            print("No label names yet!")
            return None
        else:
            return self._label_names
 

    def diagnostics_SNR(self, figname = "SNRdist.png"): 
        """ Plot SNR distributions of ref and test objects

        Parameters
        ----------
        figname: (optional) string
            title of the saved SNR diagnostic plot
        """
        print("Diagnostic for SNRs of reference and survey stars")
        data = self.tr_SNR
        plt.hist(data, bins=np.sqrt(len(data)), alpha=0.5, label="Ref Stars")
        data = self.test_SNR
        plt.hist(data, bins=np.sqrt(len(data)), alpha=0.5, label="Survey Stars")
        plt.legend(loc='upper right')
        plt.xscale('log')
        plt.title("SNR Comparison Between Reference & Test Stars")
        plt.xlabel("log(Formal SNR)")
        plt.ylabel("Number of Objects")
        plt.savefig(figname)
        plt.close()
        print("Saved fig %s" %figname)

    
    def diagnostics_ref_labels(self, figname="ref_labels_triangle.png"):
        """ Plot all training labels against each other. 
        
        Parameters
        ----------
        figname: (optional) string
            title of the saved triangle plot for reference labels
        """
        self.label_triangle_plot(self.tr_label, figname)


    def label_triangle_plot(self, label_vals, figname):
        """Make a triangle plot for the selected labels

        Parameters
        ----------
        figname: str
            if provided, save the figure into the given file

        labels: sequence
            if provided, use this sequence as text labels for each label
            dimension
        """
        labels = [r"$%s$" % l for l in self.get_plotting_labels()]
        print("Plotting every label against every other")
        fig = corner(label_vals, labels=labels, show_titles=True,
                     title_args={"fontsize":12})
        print("figname: %s" %figname)
        fig.savefig(figname)
        print("Saved fig %s" % figname)
        plt.close(fig)


    def make_contmask(self, fluxes, ivars, frac):
        """ Use training spectra to find and return continuum pixels

        For spectra split into regions, performs cont pix identification
        separately for each region.
        
        Returns
        -------
        contmask: boolean mask of length npixels
            True indicates that the pixel is continuum
        """
        print("Finding continuum pixels...")
        if self.ranges is None:
            print("assuming continuous spectra")
            contmask = find_contpix(self.wl, fluxes, ivars, frac)
        else:
            print("taking spectra in %s regions" %len(self.ranges))
            contmask = find_contpix_regions(
                    self.wl, fluxes, ivars, frac, self.ranges)
        print("%s pixels returned as continuum" %sum(contmask))
        #self.contmask = contmask
        return contmask


    def set_continuum(self, contmask):
        self.contmask = contmask


    def diagnostics_contmask(self, figname='contpix.png'):
        contmask = self.contmask
        f_bar = np.zeros(len(self.wl))
        sigma_f = np.zeros(len(self.wl))
        for wl in range(0,len(self.wl)):
            flux = self.tr_flux[:,wl]
            ivar = self.tr_ivar[:,wl]
            f_bar[wl] = np.median(flux[ivar>0])
            sigma_f[wl] = np.sqrt(np.var(flux[ivar>0]))
        bad = np.var(self.tr_ivar, axis=0) == 0
        f_bar = np.ma.array(f_bar, mask=bad)
        sigma_f = np.ma.array(sigma_f, mask=bad)
        plt.plot(self.wl, f_bar, alpha=0.7)
        plt.fill_between(self.wl, (f_bar+sigma_f), (f_bar-sigma_f), alpha=0.2)
        plt.scatter(self.wl[contmask], f_bar[contmask], c='r', label="Cont Pix")
        plt.xlabel("Wavelength (A)")
        plt.ylabel("Median Flux Across Training Objects")
        plt.legend()
        plt.title("Continuum Pix Found by The Cannon")
        plt.savefig(figname)
        print("Saving fig %s" %figname)


    def fit_continuum(self, deg, ffunc):
        if self.ranges == None:
            tr_cont = fit_cont(
                    self.tr_flux, self.tr_ivar, self.contmask, deg, ffunc)
            test_cont = fit_cont(
                    self.test_flux, self.test_ivar, self.contmask, deg, ffunc)
        else:
            tr_cont = fit_cont_regions(self.tr_flux, self.tr_ivar, 
                                       self.contmask, deg, self.ranges, ffunc)
            test_cont = fit_cont_regions(self.test_flux, self.test_ivar,
                                         self.contmask, deg, self.ranges, ffunc)
            
        return tr_cont, test_cont


    def continuum_normalize_training_q(self, q, delta_lambda):
        print("Continuum normalizing the tr set using running quantile...")
        return cont_norm_q(
            self.wl, self.tr_flux, self.tr_ivar, q=q, delta_lambda=delta_lambda)


    def continuum_normalize_f(self, cont):
        """ Continuum normalize spectra by fitting a function to continuum pix 

        For spectra split into regions, perform cont normalization
        separately for each region.
        """
        tr_cont, test_cont = cont
        if self.ranges is None:
            print("assuming continuous spectra")
            norm_tr_flux, norm_tr_ivar = cont_norm(
                    self.tr_flux, self.tr_ivar, tr_cont)
            norm_test_flux, norm_test_ivar = cont_norm(
                    self.test_flux, self.test_ivar, test_cont)
        else:
            print("taking spectra in %s regions" %(len(self.ranges)))
            norm_tr_flux, norm_tr_ivar = cont_norm_regions(
                    self.tr_flux, self.tr_ivar, tr_cont, self.ranges)
            norm_test_flux, norm_test_ivar = cont_norm_regions(
                    self.test_flux, self.test_ivar, test_cont, self.ranges)
        return norm_tr_flux, norm_tr_ivar, norm_test_flux, norm_test_ivar


    def dataset_postdiagnostics(self, figname="survey_labels_triangle.png"):
        """ Run diagnostic tests on the test set after labels have been inferred.

        Tests result in the following output: one .txt file for each label 
        listing all of the stars whose inferred labels lie >= 2 standard 
        deviations outside the reference label space, a triangle plot showing 
        all the survey labels plotted against each other, and 1-to-1 plots 
        for all of the labels showing how they compare to each other. 

        Parameters
        ----------
        """
        # Find stars whose inferred labels lie >2-sig outside ref label space
        label_names = self.get_plotting_labels()
        nlabels = len(label_names)
        reference_labels = self.tr_label
        test_labels = self.test_label_vals
        #test_IDs = np.array(self.test_IDs)
        #mean = np.mean(reference_labels, 0)
        #stdev = np.std(reference_labels, 0)
        #lower = mean - 2 * stdev
        #upper = mean + 2 * stdev
        #for i in range(nlabels):
        #    label_name = label_names[i]
        #    test_vals = test_labels[:,i]
        #    warning = np.logical_or(test_vals < lower[i], test_vals > upper[i])
        #    filename = "flagged_stars_%s.txt" % i
        #    with open(filename, 'w') as output:
        #        for star in test_IDs[warning]:
        #            output.write('{0:s}\n'.format(star))
        #    print("Reference label %s" % label_name)
        #    print("flagged %s stars beyond 2-sig of ref labels" % sum(warning))
        #    print("Saved list %s" % filename)
    
        # Plot all survey labels against each other
        figname="survey_labels_triangle.png"
        self.label_triangle_plot(self.test_label_vals, figname)
    
        # 1-1 plots of all labels
        for i in range(nlabels):
            name = self.get_plotting_labels()[i]
            orig = reference_labels[:,i]
            cannon = test_labels[:,i]
            # calculate bias and scatter
            scatter = np.round(np.std(orig-cannon),3)
            bias  = np.round(np.mean(orig-cannon),3)
            low = np.minimum(min(orig), min(cannon))
            high = np.maximum(max(orig), max(cannon))
            fig, axarr = plt.subplots(2)
            ax1 = axarr[0]
            ax1.plot([low, high], [low, high], 'k-', linewidth=2.0, label="x=y")
            ax1.scatter(orig, cannon)
            textstr = 'Scatter: %s \nBias: %s' %(scatter, bias)
            ax1.text(0.05, 0.95, textstr, transform=ax1.transAxes, fontsize=10,
                    verticalalignment='top')
            ax1.set_xlabel("Reference Value")
            ax1.set_ylabel("Cannon Output Value")
            ax1.set_title("1-1 Plot of Label " + r"$%s$" % name)
            ax2 = axarr[1]
            ax2.hist(cannon-orig, range=[-0.5,0.5])
            ax2.set_xlabel("Difference")
            ax2.set_ylabel("Count")
            ax2.set_title("Histogram of Output Minus Ref Labels")
            figname = "1to1_label_%s.png" % i
            plt.savefig(figname)
            print("Diagnostic for label output vs. input")
            print("Saved fig %s" % figname)
            plt.close()


    def set_test_label_vals(self, vals):
        """ Set label vals from an array """
        self.test_label_vals = vals





import numpy as np

LARGE = 200.
SMALL = 1. / LARGE

""" Finds and returns list of continuum pixels, as a mask. """

def find_contpix_given_cuts(f_cut, sig_cut, wl, fluxes, ivars):
    """ Find and return continuum pixels given the flux and sigma cut

    Parameters
    ----------
    f_cut: float
        the upper limit imposed on the quantity (fbar-1)
    sig_cut: float
        the upper limit imposed on the quantity (f_sig)
    wl: numpy ndarray of length npixels
        rest-frame wavelength vector
    fluxes: numpy ndarray of shape (nstars, npixels)
        pixel intensities
    ivars: numpy ndarray of shape nstars, npixels
        inverse variances, parallel to fluxes

    Returns
    -------
    contmask: boolean mask of length npixels
        True indicates that the pixel is continuum
    """
    #bad1 = np.median(ivars, axis=0) < SMALL
    bad1 = np.median(fluxes, axis=0) == 0.
    bad2 = np.var(fluxes, axis=0) == 0.
    #bad2 = np.var(ivars, axis=0) == 0
    bad = np.logical_and(bad1, bad2)
    f_bar = np.median(fluxes, axis=0)
    sigma_f = np.var(fluxes, axis=0)
    f_bar = np.ma.array(f_bar, mask=bad)
    sigma_f = np.ma.array(sigma_f, mask=bad)
    cont1 = np.abs(f_bar-1) <= f_cut
    cont2 = sigma_f <= sig_cut
    # cont3 = sigma_f >= np.abs(1-f_bar)
    contmask1 = np.logical_and(cont1, cont2)
    # contmask = np.logical_and(contmask1, cont3)
    contmask1 = np.ma.filled(contmask1, fill_value=False)
    return contmask1

def find_contpix(wl, fluxes, ivars, target_frac):
    print("Target frac: %s" %(target_frac))
    bad1 = np.median(ivars, axis=0) == SMALL
    bad2 = np.var(ivars, axis=0) == 0
    bad = np.logical_and(bad1, bad2)
    npixels = len(wl)-sum(bad)
    f_cut = 0.0001
    stepsize = 0.0001
    sig_cut = 0.0001
    contmask = find_contpix_given_cuts(f_cut, sig_cut, wl, fluxes, ivars)
    if npixels > 0:
        frac = sum(contmask)/float(npixels)
    else:
        frac = 0
    while (frac < target_frac): 
        f_cut += stepsize
        sig_cut += stepsize
        contmask = find_contpix_given_cuts(f_cut, sig_cut, wl, fluxes, ivars)
        if npixels > 0:
            frac = sum(contmask)/float(npixels)
        else:
            frac = 0
    if frac > 0.10*npixels:
        print("Warning: Over 10% of pixels identified as continuum.")
    print("%s out of %s pixels identified as continuum" %(sum(contmask), 
                                                          npixels))
    print("Cuts: f_cut %s, sig_cut %s" %(f_cut, sig_cut))
    return contmask

def find_contpix_regions(wl, fluxes, ivars, frac, ranges):
    print("taking spectra in %s regions" %len(ranges))
    contmask = np.zeros(len(wl), dtype=bool)
    for chunk in ranges:
        start = chunk[0]
        stop = chunk[1]
        contmask[start:stop] = find_contpix(
                wl[start:stop], fluxes[:,start:stop], ivars[:,start:stop], frac)
    return contmask



__all__ = ["corner", "hist2d", "error_ellipse"]
__version__ = "0.1.1"
__author__ = "Dan Foreman-Mackey (danfm@nyu.edu)"
__copyright__ = "Copyright 2013 Daniel Foreman-Mackey"
__contributors__ = [
    # Alphabetical by first name.
    "Adrian Price-Whelan @adrn",
    "Brendon Brewer @eggplantbren",
    "Ekta Patel @ekta1224",
    "Emily Rice @emilurice",
    "Geoff Ryan @geoffryan",
    "Kyle Barbary @kbarbary",
    "Phil Marshall @drphilmarshall",
    "Pierre Gratier @pirg",
]

import numpy as np
import matplotlib.pyplot as pl
from matplotlib.ticker import MaxNLocator
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.patches import Ellipse
import matplotlib.cm as cm


def corner(xs, weights=None, labels=None, show_titles=False, title_fmt=".2f",
           title_args=None, extents=None, truths=None, truth_color="#4682b4",
           scale_hist=False, quantiles=None, verbose=True,
           plot_contours=True, plot_datapoints=True, fig=None, **kwargs):
    """
    Make a *sick* corner plot showing the projections of a data set in a
    multi-dimensional space. kwargs are passed to hist2d() or used for
    `matplotlib` styling.

    Parameters
    ----------
    xs : array_like (nsamples, ndim)
        The samples. This should be a 1- or 2-dimensional array. For a 1-D
        array this results in a simple histogram. For a 2-D array, the zeroth
        axis is the list of samples and the next axis are the dimensions of
        the space.

    weights : array_like (nsamples,)
        The weight of each sample. If `None` (default), samples are given
        equal weight.

    labels : iterable (ndim,) (optional)
        A list of names for the dimensions.

    show_titles : bool (optional)
        Displays a title above each 1-D histogram showing the 0.5 quantile
        with the upper and lower errors supplied by the quantiles argument.

    title_fmt : string (optional)
        The format string for the quantiles given in titles.
        (default: `.2f`)

    title_args : dict (optional)
        Any extra keyword arguments to send to the `add_title` command.

    extents : iterable (ndim,) (optional)
        A list where each element is either a length 2 tuple containing
        lower and upper bounds (extents) or a float in range (0., 1.)
        giving the fraction of samples to include in bounds, e.g.,
        [(0.,10.), (1.,5), 0.999, etc.].
        If a fraction, the bounds are chosen to be equal-tailed.

    truths : iterable (ndim,) (optional)
        A list of reference values to indicate on the plots.

    truth_color : str (optional)
        A ``matplotlib`` style color for the ``truths`` makers.

    scale_hist : bool (optional)
        Should the 1-D histograms be scaled in such a way that the zero line
        is visible?

    quantiles : iterable (optional)
        A list of fractional quantiles to show on the 1-D histograms as
        vertical dashed lines.

    verbose : bool (optional)
        If true, print the values of the computed quantiles.

    plot_contours : bool (optional)
        Draw contours for dense regions of the plot.

    plot_datapoints : bool (optional)
        Draw the individual data points.

    fig : matplotlib.Figure (optional)
        Overplot onto the provided figure object.

    """
    if quantiles is None:
        quantiles = []

    if title_args is None:
        title_args = {}

    # Deal with 1D sample lists.
    xs = np.atleast_1d(xs)
    if len(xs.shape) == 1:
        xs = np.atleast_2d(xs)
    else:
        assert len(xs.shape) == 2, "The input sample array must be 1- or 2-D."
        xs = xs.T
    assert xs.shape[0] <= xs.shape[1], "I don't believe that you want more " \
                                       "dimensions than samples!"

    if weights is not None:
        weights = np.asarray(weights)
        if weights.ndim != 1:
            raise ValueError('weights must be 1-D')
        if xs.shape[1] != weights.shape[0]:
            raise ValueError('lengths of weights must match number of samples')

    # backwards-compatibility
    plot_contours = kwargs.get("smooth", plot_contours)

    K = len(xs)
    factor = 2.0           # size of one side of one panel
    lbdim = 0.5 * factor   # size of left/bottom margin
    trdim = 0.2 * factor   # size of top/right margin
    whspace = 0.05         # w/hspace size
    plotdim = factor * K + factor * (K - 1.) * whspace
    dim = lbdim + plotdim + trdim

    if fig is None:
        fig, axes = pl.subplots(K, K, figsize=(dim, dim))
    else:
        try:
            axes = np.array(fig.axes).reshape((K, K))
        except:
            raise ValueError("Provided figure has {0} axes, but data has "
                             "dimensions K={1}".format(len(fig.axes), K))
    lb = lbdim / dim
    tr = (lbdim + plotdim) / dim
    fig.subplots_adjust(left=lb, bottom=lb, right=tr, top=tr,
                        wspace=whspace, hspace=whspace)

    if extents is None:
        extents = [[x.min(), x.max()] for x in xs]

        # Check for parameters that never change.
        m = np.array([e[0] == e[1] for e in extents], dtype=bool)
        if np.any(m):
            raise ValueError(("It looks like the parameter(s) in column(s) "
                              "{0} have no dynamic range. Please provide an "
                              "`extent` argument.")
                             .format(", ".join(map("{0}".format,
                                                   np.arange(len(m))[m]))))
    else:
        # If any of the extents are percentiles, convert them to ranges.
        for i in range(len(extents)):
            try:
                emin, emax = extents[i]
            except TypeError:
                q = [0.5 - 0.5*extents[i], 0.5 + 0.5*extents[i]]
                extents[i] = quantile(xs[i], q, weights=weights)

    for i, x in enumerate(xs):
        if np.shape(xs)[0] == 1:
            ax = axes
        else:
            ax = axes[i, i]
        # Plot the histograms.
        n, b, p = ax.hist(x, weights=weights, bins=kwargs.get("bins", 50),
                          range=extents[i], histtype="step",
                          color=kwargs.get("color", "k"))
        if truths is not None:
            ax.axvline(truths[i], color=truth_color)

        # Plot quantiles if wanted.
        if len(quantiles) > 0:
            qvalues = quantile(x, quantiles, weights=weights)
            for q in qvalues:
                ax.axvline(q, ls="dashed", color=kwargs.get("color", "k"))

            if verbose:
                print("Quantiles:")
                print([item for item in zip(quantiles, qvalues)])

            if show_titles:
                # Compute the quantiles for the title. This might redo
                # unneeded computation but who cares.
                q_16, q_50, q_84 = quantile(x, [0.16, 0.5, 0.84],
                                            weights=weights)
                q_m, q_p = q_50-q_16, q_84-q_50

                # Format the quantile display.
                fmt = "{{0:{0}}}".format(title_fmt).format
                title = r"${{{0}}}_{{-{1}}}^{{+{2}}}$"
                title = title.format(fmt(q_50), fmt(q_m), fmt(q_p))

                # Add in the column name if it's given.
                if labels is not None:
                    title = "{0} = {1}".format(labels[i], title)

                # Add the title to the axis.
                ax.set_title(title, **title_args)

        # Set up the axes.
        ax.set_xlim(extents[i])
        if scale_hist:
            maxn = np.max(n)
            ax.set_ylim(-0.1 * maxn, 1.1 * maxn)
        else:
            ax.set_ylim(0, 1.1 * np.max(n))
        ax.set_yticklabels([])
        ax.xaxis.set_major_locator(MaxNLocator(5))

        # Not so DRY.
        if i < K - 1:
            ax.set_xticklabels([])
        else:
            [l.set_rotation(45) for l in ax.get_xticklabels()]
            if labels is not None:
                ax.set_xlabel(labels[i])
                ax.xaxis.set_label_coords(0.5, -0.3)

        for j, y in enumerate(xs):
            if np.shape(xs)[0] == 1:
                ax = axes
            else:
                ax = axes[i, j]
            if j > i:
                ax.set_visible(False)
                ax.set_frame_on(False)
                continue
            elif j == i:
                continue

            hist2d(y, x, ax=ax, extent=[extents[j], extents[i]],
                   plot_contours=plot_contours,
                   plot_datapoints=plot_datapoints,
                   weights=weights, **kwargs)

            if truths is not None:
                ax.plot(truths[j], truths[i], "s", color=truth_color)
                ax.axvline(truths[j], color=truth_color)
                ax.axhline(truths[i], color=truth_color)

            ax.xaxis.set_major_locator(MaxNLocator(5))
            ax.yaxis.set_major_locator(MaxNLocator(5))

            if i < K - 1:
                ax.set_xticklabels([])
            else:
                [l.set_rotation(45) for l in ax.get_xticklabels()]
                if labels is not None:
                    ax.set_xlabel(labels[j])
                    ax.xaxis.set_label_coords(0.5, -0.3)

            if j > 0:
                ax.set_yticklabels([])
            else:
                [l.set_rotation(45) for l in ax.get_yticklabels()]
                if labels is not None:
                    ax.set_ylabel(labels[i])
                    ax.yaxis.set_label_coords(-0.3, 0.5)

    return fig


def quantile(x, q, weights=None):
    """
    Like numpy.percentile, but:

    * Values of q are quantiles [0., 1.] rather than percentiles [0., 100.]
    * scalar q not supported (q must be iterable)
    * optional weights on x

    """
    if weights is None:
        return np.percentile(x, [100. * qi for qi in q])
    else:
        idx = np.argsort(x)
        xsorted = x[idx]
        cdf = np.add.accumulate(weights[idx])
        cdf /= cdf[-1]
        return np.interp(q, cdf, xsorted).tolist()


def error_ellipse(mu, cov, ax=None, factor=1.0, **kwargs):
    """
    Plot the error ellipse at a point given its covariance matrix.

    """
    # some sane defaults
    facecolor = kwargs.pop('facecolor', 'none')
    edgecolor = kwargs.pop('edgecolor', 'k')

    x, y = mu
    U, S, V = np.linalg.svd(cov)
    theta = np.degrees(np.arctan2(U[1, 0], U[0, 0]))
    ellipsePlot = Ellipse(xy=[x, y],
                          width=2 * np.sqrt(S[0]) * factor,
                          height=2 * np.sqrt(S[1]) * factor,
                          angle=theta,
                          facecolor=facecolor, edgecolor=edgecolor, **kwargs)

    if ax is None:
        ax = pl.gca()
    ax.add_patch(ellipsePlot)

    return ellipsePlot


def hist2d(x, y, *args, **kwargs):
    """
    Plot a 2-D histogram of samples.

    """
    ax = kwargs.pop("ax", pl.gca())

    extent = kwargs.pop("extent", [[x.min(), x.max()], [y.min(), y.max()]])
    bins = kwargs.pop("bins", 50)
    color = kwargs.pop("color", "k")
    linewidths = kwargs.pop("linewidths", None)
    plot_datapoints = kwargs.get("plot_datapoints", True)
    plot_contours = kwargs.get("plot_contours", True)

    cmap = cm.get_cmap("gray")
    cmap._init()
    cmap._lut[:-3, :-1] = 0.
    cmap._lut[:-3, -1] = np.linspace(1, 0, cmap.N)

    X = np.linspace(extent[0][0], extent[0][1], bins + 1)
    Y = np.linspace(extent[1][0], extent[1][1], bins + 1)
    try:
        H, X, Y = np.histogram2d(x.flatten(), y.flatten(), bins=(X, Y),
                                 weights=kwargs.get('weights', None))
    except ValueError:
        raise ValueError("It looks like at least one of your sample columns "
                         "have no dynamic range. You could try using the "
                         "`extent` argument.")

    V = 1.0 - np.exp(-0.5 * np.arange(0.5, 2.1, 0.5) ** 2)
    Hflat = H.flatten()
    inds = np.argsort(Hflat)[::-1]
    Hflat = Hflat[inds]
    sm = np.cumsum(Hflat)
    sm /= sm[-1]

    for i, v0 in enumerate(V):
        try:
            V[i] = Hflat[sm <= v0][-1]
        except:
            V[i] = Hflat[0]

    X1, Y1 = 0.5 * (X[1:] + X[:-1]), 0.5 * (Y[1:] + Y[:-1])
    X, Y = X[:-1], Y[:-1]

    if plot_datapoints:
        ax.plot(x, y, "o", color=color, ms=1.5, zorder=-1, alpha=0.1,
                rasterized=True)
        if plot_contours:
            ax.contourf(X1, Y1, H.T, [V[-1], H.max()],
                        cmap=LinearSegmentedColormap.from_list("cmap",
                                                               ([1] * 3,
                                                                [1] * 3),
                        N=2), antialiased=False)

    if plot_contours:
        ax.pcolor(X, Y, H.max() - H.T, cmap=cmap)
        ax.contour(X1, Y1, H.T, V, colors=color, linewidths=linewidths)

    data = np.vstack([x, y])
    mu = np.mean(data, axis=1)
    cov = np.cov(data)
    if kwargs.pop("plot_ellipse", False):
        error_ellipse(mu, cov, ax=ax, edgecolor="r", ls="dashed")

    ax.set_xlim(extent[0])
    ax.set_ylim(extent[1])



# Data munging for RAVE

# The Cannon needs: 
# length-L wavelength vec 
# an NxL block of training set pixel fluxes and corresponding ivars
# an NxK block of training label values
# an MxL block of test set pixel vals and corresponding ivars

def read_wl():
    filename = ('2013K1_parameters.save')
    items = inputf.items()
    data = items[0][1]
    wl = data['lambda'][0] # assuming they're all the same... 
    return wl

def read_tr_data():
    inputf = readsav('RAVE_DR4_calibration_data.save')
    items = inputf.items()
    data = items[0][1]
    tr_flux = data['spectrum'][0].T # shape (807, 839) = (nstars, npix)
    npix = tr_flux.shape[1]
    nstars = tr_flux.shape[0]
    teff = data['teff'][0] # length 807
    logg = data['logg'][0] # length 807
    feh = data['feh'][0]
    tr_label = np.vstack((teff, logg, feh)).T
    snr = np.zeros(nstars)
    snr.fill(100) # a guess for what the SNR could be
    tr_ivar = (snr[:,None]/tr_flux)**2
    return tr_flux, tr_ivar, tr_label

def read_test(filename):
    inputf = readsav(filename)
    items = inputf.items()
    data = items[0][1]
    sp = data['obs_sp'] # (75437, 839) 
    test_flux = np.zeros((len(sp), len(sp[0])))
    for jj in range(0, len(sp)):
        test_flux[jj,:] = sp[jj]
    snr = np.array(data['snr'])
    test_ivar = (snr[:,None]/test_flux)**2
    bad = np.logical_or(np.isnan(test_ivar), np.isnan(test_flux))
    test_ivar[bad] = 0.
    test_flux[bad] = 0.
    return (test_flux, test_ivar, wl)




"""This runs Step 2 of The Cannon:
    uses the model to solve for the labels of the test set."""

from __future__ import (absolute_import, division, print_function, unicode_literals)

from scipy import optimize as opt
import numpy as np

LARGE = 200.
SMALL = 1. / LARGE

def get_lvec(labels):
    """
    Constructs a label vector for an arbitrary number of labels
    Assumes that our model is quadratic in the labels

    Parameters
    ----------
    labels: numpy ndarray
        pivoted label values for one star

    Returns
    -------
    lvec: numpy ndarray
        label vector
    """
    nlabels = len(labels)
    # specialized to second-order model
    linear_terms = labels
    quadratic_terms = np.outer(linear_terms, 
                               linear_terms)[np.triu_indices(nlabels)]
    lvec = np.hstack((linear_terms, quadratic_terms))
    return lvec


def func(coeffs, *labels):
    """ Takes the dot product of coefficients vec & labels vector 
    
    Parameters
    ----------
    coeffs: 

    *labels: 

    Returns
    -------
    dot product of coeffs vec and labels vec
    
    """
    lvec = get_lvec(list(labels))
    return np.dot(coeffs, lvec)


def infer_labels(model, dataset):
    """
    Uses the model to solve for labels of the test set.

    Parameters
    ----------
    model: tuple
        coeffs_all, covs, scatters, chis, chisqs, pivots
        result from :func:`train_model`

    test_set: Dataset
        dataset that needs label inference

    Returns
    -------
    test_set: Dataset
        same dataset as the input value with updated label_vals attribute

    covs_all:
        covariance matrix of the fit
    """
    print("Inferring Labels...")
    coeffs_all, covs, scatters, red_chisqs, pivots, label_vector = model
    nlabels = len(pivots)
    fluxes = dataset.test_flux
    ivars = dataset.test_ivar
    nstars = fluxes.shape[0]
    labels_all = np.zeros((nstars, nlabels))
    MCM_rotate_all = np.zeros((nstars, coeffs_all.shape[1] - 1,
                               coeffs_all.shape[1]-1.))
    #covs_all = np.zeros((nstars, nlabels, nlabels))
    errs_all = np.zeros((nstars, nlabels))

    for jj in range(nstars):
        flux = fluxes[jj,:]
        ivar = ivars[jj,:]
        flux_piv = flux - coeffs_all[:,0] * 1.  # pivot around the leading term
        #Cinv = ivar / (1 + ivar * scatters**2)
        #bad = ivar == SMALL
        #sig = np.zeros(ivar.shape)
        #sig = np.ma.array(sig, mask=bad)
        #ivar = np.ma.array(ivar, mask=bad)
        #scatters = np.ma.array(scatters, mask=bad)
        sig = np.sqrt(1./ivar + scatters**2)
        coeffs = np.delete(coeffs_all, 0, axis=1)  # take pivot into account
        #print(sig)
        try:
            labels, covs = opt.curve_fit(func, coeffs, flux_piv,
                                         p0=np.repeat(1, nlabels),
                                         sigma=sig, absolute_sigma=True)
        except TypeError:  # old scipy version
            labels, covs = opt.curve_fit(func, coeffs, flux_piv,
                                         p0=np.repeat(1, nlabels), sigma=sig)
            # rescale covariance matrix
            chi = (flux_piv-func(coeffs, *labels)) / sig
            chi2 = (chi**2).sum()
            # FIXME: dof does not seem to be right to me (MF)
            dof = len(flux_piv) - nlabels
            factor = (chi2 / dof)
            covs /= factor
        labels = labels + pivots
        #MCM_rotate = np.dot(coeffs.T, Cinv[:,None] * coeffs)
        labels_all[jj,:] = labels
        #MCM_rotate_all[jj, :, :] = MCM_rotate
        errs_all[jj,:] = covs.diagonal()
        #covs_all[jj, :, :] = covs

    dataset.set_test_label_vals(labels_all)
    return dataset, errs_all



from __future__ import (absolute_import, division, print_function)
from apogee import load_spectra, load_labels
from model import CannonModel
from dataset import Dataset
from spectral_model import draw_spectra, diagnostics, triangle_pixels, overlay_spectra, residuals
import numpy as np

# (1) PREPARE DATA

wl, tr_flux, tr_ivar = load_spectra("example_DR10/Data")
test_flux = tr_flux
test_ivar = tr_ivar
all_labels = load_labels("example_DR10/reference_labels.csv")
teff_corr = all_labels[:,1]
logg_corr = all_labels[:,3]
mh_corr = all_labels[:,5]
tr_label = np.vstack((teff_corr, logg_corr, mh_corr)).T
dataset = Dataset(wl, tr_flux, tr_ivar, tr_label, test_flux, test_ivar)
dataset.ranges = [[371,3192], [3697,5997], [6461,8255]]

# optional: set headers for plotting
dataset.set_label_names(['T_{eff}', '\log g', '[M/H]'])

# Plot SNR distributions and triangle plot of reference labels
dataset.diagnostics_SNR()
dataset.diagnostics_ref_labels()

# (2) IDENTIFY CONTINUUM PIXELS

pseudo_tr_flux, pseudo_tr_ivar = dataset.continuum_normalize_training_q(
        q=0.90, delta_lambda=50)

# in each region of the pseudo cont normed tr spectrum, 
# identify the best 7% of continuum pix
contmask = dataset.make_contmask(pseudo_tr_flux, pseudo_tr_ivar, frac=0.07)
dataset.set_continuum(contmask)
cont = dataset.fit_continuum(3, "sinusoid")

# (3) RUN CONTINUUM NORMALIZATION CODE
norm_tr_flux, norm_tr_ivar, norm_test_flux, norm_test_ivar = \
        dataset.continuum_normalize_f(cont)

dataset.tr_flux = norm_tr_flux
dataset.tr_ivar = norm_tr_ivar
dataset.test_flux = norm_test_flux
dataset.test_ivar = norm_test_ivar

# (4) TRAINING STEP

# learn the model from the reference_set
model = CannonModel(dataset, 2) # 2 = quadratic model
model.fit() # model.train would work equivalently.

# check the model
model.diagnostics()

# (5) TEST STEP

# infer labels with the new model for the test_set
dataset, label_errs = model.infer_labels(dataset)

# Make plots
dataset.dataset_postdiagnostics(dataset)

cannon_set = model.draw_spectra(dataset)
# model.spectral_diagnostics(dataset)



# This is a demonstration of how to run The Cannon on RAVE Data

import numpy as np
import pickle
import glob
from scipy.io.idl import readsav

from cannon.dataset import Dataset
from cannon.model import CannonModel


# STEP 1: PREPARE DATA

# The Cannon needs: length-L wavelength vec 
# an NxL block of training set pixel fluxes and corresponding ivars
# an NxK block of training label values
# an MxL block of test set pixel vals and corresponding ivars

# training set
inputf = readsav('RAVE_DR4_calibration_data.save') 
items = inputf.items()
data = items[0][1]

tr_flux = data['spectrum'][0].T # shape (807, 839) = (nstars, npix)
npix = tr_flux.shape[1]
nstars = tr_flux.shape[0]

teff = data['teff'][0] # length 807
logg = data['logg'][0] # length 807
feh = data['feh'][0]
tr_label = np.vstack((teff, logg, feh)).T

snr = np.zeros(nstars)
snr.fill(100) # a guess for what the SNR could be
tr_ivar = (snr[:,None]/tr_flux)**2

# test set
def read_test(filename): 
    inputf = readsav(filename)
    items = inputf.items()
    data = items[0][1]
    sp = data['obs_sp'] # (75437, 839) 
    test_flux = np.zeros((len(sp), len(sp[0])))
    for jj in range(0, len(sp)):
        test_flux[jj,:] = sp[jj]
    snr = np.array(data['snr'])
    test_ivar = (snr[:,None]/test_flux)**2
    bad = np.logical_or(np.isnan(test_ivar), np.isnan(test_flux)) 
    test_ivar[bad] = 0.
    test_flux[bad] = 0.
    return (test_flux, test_ivar, wl)

if glob.glob('rave_test_data.p'):
    (wl, test_flux, test_ivar) = pickle.load(
            open('rave_test_data.p', 'r'))

else:
    a = read_test('2009K1_parameters.save')
    b = read_test('2013K1_parameters.save')

    test_flux = np.vstack((a[0], b[0]))
    test_ivar = np.vstack((a[1], b[1]))
    wl = a[2] # same for both sets
    pickle.dump((wl, test_flux, test_ivar), open("rave_test_data.p", "w"), -1)

wl = pickle.load(open("rave_wl.p", "r"))

# initialize a dataset object
dataset = Dataset(wl, tr_flux, tr_ivar, tr_label, test_flux, test_ivar)

# diagnostic plots for input spectra and reference labels
dataset.set_label_names(['T_{eff}', '\log g', '[Fe/H]'])
dataset.diagnostics_SNR()
dataset.diagnostics_ref_labels()


# STEP 2: CONTINUUM IDENTIFICATION

# Identify continuum pixels using a median and var flux cut
# Split spectrum into three regions, to make it more evenly spaced
dataset.ranges = [[0,200], [200,400], [400,600], [600,len(dataset.wl)]]
contmask = dataset.make_contmask(dataset.tr_flux, dataset.tr_ivar, frac=0.05)
dataset.set_continuum(contmask)
dataset.diagnostics_contmask()


# STEP 3: CONTINUUM NORMALIZATION
dataset.ranges = None
if glob.glob('rave_cont.p'):
    (tr_cont, test_cont) = pickle.load(open('rave_cont.p', 'r'))
else:
    tr_cont, test_cont = dataset.fit_continuum(deg=3, ffunc="sinusoid")
    pickle.dump((tr_cont, test_cont), open("rave_cont.p", "w"))

norm_tr_flux, norm_tr_ivar, norm_test_flux, norm_test_ivar = \
        dataset.continuum_normalize_f(cont=(tr_cont, test_cont))

dataset.tr_flux = norm_tr_flux
dataset.tr_ivar = norm_tr_ivar
dataset.test_flux = norm_test_flux
dataset.test_ivar = norm_test_ivar

model = CannonModel(dataset, 2)
model.fit()
model.diagnostics()
dataset, label_errs = model.infer_labels(dataset)
dataset.dataset_postdiagnostics(dataset)




from setuptools import setup

setup(name='TheCannon',
        version='0.2.7',
        description='Data-driven stellar parameters and abundances from spectra',
        url='http://github.com/annayqho/TheCannon',
        author='Anna Y. Q. Ho',
        author_email='annayqho@gmail.com',
        license='MIT',
        packages=['TheCannon'],
        )



from __future__ import (absolute_import, division, print_function,)
import numpy as np
import scipy.optimize as opt
import os
import sys
from .helpers import Table
import matplotlib.pyplot as plt

# python 3 special
PY3 = sys.version_info[0] > 2
if not PY3:
    range = xrange

try:
    from astropy.io import fits as pyfits
except ImportError:
    import pyfits

def get_pixmask(fluxes, flux_errs):
    """ Return a mask array of bad pixels for one object's spectrum

    Bad pixels are defined as follows: fluxes or errors are not finite, or 
    reported errors are negative, or the standard deviation of the fluxes
    across all the stars is zero (if that pixel is exactly the same, then
    we're looking at the gaps in the spectrum.)

    Parameters
    ----------
    fluxes: ndarray
        flux array

    flux_errs: ndarray
        measurement uncertainties on fluxes

    Returns
    -------
    mask: ndarray, dtype=bool
        array giving bad pixels as True values
    """
    bad_flux = (~np.isfinite(fluxes)) 
    bad_err = (~np.isfinite(flux_errs)) | (flux_errs <= 0)
    bad_pix = bad_err | bad_flux

    return bad_pix

def load_spectra(data_dir):
    """
    Extracts spectra (wavelengths, fluxes, fluxerrs) from apogee fits files

    Returns
    -------
    IDs: list of length nstars
        stellar IDs
    
    wl: numpy ndarray of length npixels
        rest-frame wavelength vector

    fluxes: numpy ndarray of shape (nstars, npixels)
        training set or test set pixel intensities

    ivars: numpy ndarray of shape (nstars, npixels)
        inverse variances, parallel to fluxes
        
    SNRs: numpy ndarray of length nstars
    """
    print("Loading spectra from directory %s" %data_dir)
    files = list(sorted([data_dir + "/" + filename
             for filename in os.listdir(data_dir)]))
    nstars = len(files)  
    
    for jj, fits_file in enumerate(files):
        file_in = pyfits.open(fits_file)
        flux = np.array(file_in[1].data)
        if jj == 0:
            npixels = len(flux)
            fluxes = np.zeros((nstars, npixels), dtype=float)
            ivars = np.zeros(fluxes.shape, dtype=float)
            start_wl = file_in[1].header['CRVAL1']
            diff_wl = file_in[1].header['CDELT1']
            val = diff_wl * (npixels) + start_wl
            wl_full_log = np.arange(start_wl,val, diff_wl)
            wl_full = [10 ** aval for aval in wl_full_log]
            wl = np.array(wl_full)
        flux_err = np.array((file_in[2].data))
        badpix = get_pixmask(flux, flux_err)
        flux = np.ma.array(flux, mask=badpix)
        flux_err = np.ma.array(flux_err, mask=badpix)
        ones = np.ma.array(np.ones(npixels), mask=badpix)
        ivar = ones / flux_err**2
        ivar = np.ma.filled(ivar, fill_value=0.)
        fluxes[jj,:] = flux
        ivars[jj,:] = ivar

    print("Spectra loaded")
    return wl, fluxes, ivars

def load_labels(filename):
    print("Loading reference labels from file %s" %filename)
    data = Table(filename)
    data.sort('id')
    label_names = data.keys()[1:] # ignore id
    nlabels = len(label_names)
    print('%s labels:' %nlabels)
    print(label_names)
    labels = np.array([data[k] for k in label_names], dtype=float).T
    return labels 
